""" Remark Qualifier """
from rsa_835_parser.elements import Element, Code

remark_qualifiers = {
	'HE': 'claim payment'
}


class RemarkQualifier(Element):
    """ Remark Qualifier class """

    def parser(self, value: str) -> Code:
        description = remark_qualifiers.get(value, None)
        return Code(value, description)



""" Integer Element """
from typing import Optional, Union

from rsa_835_parser.elements import Element


class Integer(Element):
    """ Integer class """

    def parser(self, value: str) -> Optional[Union[int, str]]:
        if value == '':
            return None

        try:
            value = int(value)
        except ValueError:
            pass

        return value



""" Authorization Information Qualifier """
from typing import Optional

from rsa_835_parser.elements import Element


class AuthorizationInformationQualifier(Element):
    """Authorization Information Qualifier class"""

    def parser(self, value: str) -> Optional[str]:
        if value == "00":
            value = None

        return value



""" Service segment """
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.elements.dollars import Dollars
from rsa_835_parser.elements.service_code import ServiceCode
from rsa_835_parser.elements.service_qualifier import ServiceQualifier
from rsa_835_parser.elements.service_modifier import ServiceModifier
from rsa_835_parser.elements.integer import Integer
from rsa_835_parser.segments.utilities import split_segment, get_element

class Service:
    """Service segment"""

    identification = "SVC"

    identifier = Identifier()
    charge_amount = Dollars()
    paid_amount = Dollars()
    code = ServiceCode()
    qualifier = ServiceQualifier()
    modifier = ServiceModifier()
    allowed_units = Integer()
    billed_units = Integer()

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.code = segment[1]
        self.qualifier = segment[1]
        self.modifier = segment[1]
        self.charge_amount = segment[2]
        self.paid_amount = segment[3]

        # assume unit count of one if unit not provided
        default = 0 if self.paid_amount == 0 else 1
        self.allowed_units = get_element(segment, 5, default=default)

        self.billed_units = get_element(segment, 7, default=self.allowed_units)

    def __repr__(self):
        return "\n".join(str(item) for item in self.__dict__.items())


if __name__ == "__main__":
    pass



""" Organization Element """
from rsa_835_parser.elements import Element

organizations = {
	'AV09311993': 'Availity',
	'ZIRMED': 'Zirmed'
}

class Organization(Element):
    """ Organization class """

    def parser(self, value: str) -> str:
        value = value.strip()
        return organizations.get(value, value)



""" TransactionSets class definition. """

# Standard library imports
from typing import List, Iterable

# Third party imports
import pandas as pd

# Local application imports
from rsa_835_parser.transaction_set.transaction_set import TransactionSet


class TransactionSets:
    """A collection of TransactionSet objects."""

    def __init__(self, transaction_sets: List[TransactionSet]):
        self.transaction_sets = transaction_sets

    def __iter__(self) -> Iterable[TransactionSet]:
        yield from self.transaction_sets

    def __len__(self) -> int:
        return len(self.transaction_sets)

    def __repr__(self):
        return "\n".join(str(transaction_set) for transaction_set in self)

    def to_dataframe(self) -> pd.DataFrame:
        """flatten the remittance advice by service to a pandas DataFrame"""
        data = pd.DataFrame()
        for transaction_set in self:
            data = pd.concat([data, transaction_set.to_dataframe()])

            data = TransactionSets.sort_columns(data)
            return data

    @staticmethod
    def sort_columns(data: pd.DataFrame) -> pd.DataFrame:
        """Sort columns in a pandas DataFrame."""
        substrings = ["adj", "ref", "rem"]
        variable_columns = [
            c for c in data.columns if any(sub_string in c for sub_string in substrings)
        ]
        variable_columns = sorted(variable_columns)

        static_columns = [c for c in data.columns if c not in variable_columns]

        data = data[static_columns + variable_columns]
        return data

    def sum_payments(self) -> float:
        """Sum the total amount paid in the remittance advice."""
        amount = 0
        for transaction_set in self:
            amount += transaction_set.financial_information.amount_paid

        return amount

    def count_claims(self) -> int:
        """Count the number of claims in the remittance advice."""
        count = 0
        for transaction_set in self:
            count += len(transaction_set.claims)

        return count

    def count_patients(self) -> int:
        """Count the number of patients in the remittance advice."""
        patients = []
        for transaction_set in self:
            for claim in transaction_set.claims:
                patient = claim.patient
                patients.append(patient.identification_code)

        patients = set(patients)
        return len(patients)



""" TransactionSet class for parsing EDI 835 remittance advices """

# Standard library imports
from typing import List, Iterator, Optional
from collections import namedtuple
from idna import check_label

# Third party imports
import pandas as pd

# Local application imports
from rsa_835_parser.loops.claim import Claim as ClaimLoop
from rsa_835_parser.loops.service import Service as ServiceLoop
from rsa_835_parser.loops.organization import Organization as OrganizationLoop
from rsa_835_parser.segments.utilities import find_identifier
from rsa_835_parser.segments.interchange import Interchange as InterchangeSegment
from rsa_835_parser.segments.eob_information import CheckEFT
from rsa_835_parser.segments.financial_information import (
    FinancialInformation as FinancialInformationSegment,
)

BuildAttributeResponse = namedtuple(
    "BuildAttributeResponse", "key value segment segments"
)


class TransactionSet:
    """TransactionSet class for parsing EDI 835 remittance advices"""

    def __init__(
        self,
        check_information: CheckEFT,
        interchange: InterchangeSegment,
        financial_information: FinancialInformationSegment,
        claims: List[ClaimLoop],
        organizations: List[OrganizationLoop],
        file_path: str = None,
    ):
        self.check_information = check_information
        self.interchange = interchange
        self.financial_information = financial_information
        self.claims = claims if claims else []
        self.organizations = organizations if organizations else []
        self.file_path = file_path

    def __repr__(self):
        return "\n".join(str(item) for item in self.__dict__.items())

    @property
    def payer(self) -> OrganizationLoop:
        """return the payer organization"""
        payer = [o for o in self.organizations if o.organization.type == "payer"]
        assert len(payer) == 1
        return payer[0]

    @property
    def payee(self) -> OrganizationLoop:
        """return the payee organization"""
        payee = [o for o in self.organizations if o.organization.type == "payee"]
        assert len(payee) == 1
        return payee[0]

    def to_dataframe(self) -> pd.DataFrame:
        """flatten the remittance advice by service to a pandas DataFrame"""
        data = []

        for claim in self.claims:
            for service in claim.services:
                datum = TransactionSet.serialize_service(
                    self.check_information, self.financial_information, self.payer, claim, service
                )

                for index, adjustment in enumerate(service.adjustments):
                    datum[f"adj_{index}_group"] = adjustment.group_code.code
                    datum[f"adj_{index}_code"] = adjustment.reason_code.code
                    datum[f"adj_{index}_amount"] = adjustment.amount

                for index, reference in enumerate(service.references):
                    datum[f"ref_{index}_qual"] = reference.qualifier.code
                    datum[f"ref_{index}_value"] = reference.value

                for index, remark in enumerate(service.remarks):
                    datum[f"rem_{index}_qual"] = remark.qualifier.code
                    datum[f"rem_{index}_code"] = remark.code.code

                data.append(datum)

        return pd.DataFrame(data)

    @staticmethod
    def serialize_service(
        check_information: CheckEFT,
        financial_information: FinancialInformationSegment,
        payer: OrganizationLoop,
        claim: ClaimLoop,
        service: ServiceLoop,
    ) -> dict:
        """serialize a service to a dictionary"""

        # if the service doesn't have a start date assume the service and claim dates match
        start_date = None
        if service.service_period_start:
            start_date = service.service_period_start.date
        elif claim.claim_statement_period_start:
            start_date = claim.claim_statement_period_start.date

        # if the service doesn't have an end date assume the service and claim dates match
        end_date = None
        if service.service_period_end:
            end_date = service.service_period_end.date
        elif claim.claim_statement_period_end:
            end_date = claim.claim_statement_period_end.date

        datum = {
            "marker": claim.claim.marker,
            "claim_number": claim.claim.claim_number,
            "patient": claim.patient.last_first_name,
            "code": service.service.code,
            "modifier": service.service.modifier,
            "qualifier": service.service.qualifier,
            "allowed_units": service.service.allowed_units,
            "billed_units": service.service.billed_units,
            "check_number": check_information.check_number,
            "transaction_date": financial_information.transaction_date,
            "charge_amount": service.service.charge_amount,
            "allowed_amount": service.allowed_amount,
            "paid_amount": service.service.paid_amount,
            "payer": payer.organization.name,
            "start_date": start_date,
            "end_date": end_date,
            "rendering_provider": claim.rendering_provider.name
            if claim.rendering_provider
            else None,
            "payer_classification": str(claim.claim.status.payer_classification),
        }

        return datum

    @staticmethod
    def serialize_claim(
        check_information: CheckEFT,
        financial_information: FinancialInformationSegment,
        payer: OrganizationLoop,
        claim: ClaimLoop,
    ) -> dict:
        """serialize a Claim Summary Data to a dictionary"""

        # Pending to Complete
        # --------------------
        datum = {
            "Service Dates": claim.claim.service_period_start.date + " - " + claim.claim.service_period_end.date,
            "Claim Number": claim.claim.claim_number,
            "Payer": payer.organization.name,
            "Check/EFT Number (Check/EFT Date)": check_information.check_number + ' (' + financial_information.transaction_date + ')',
            "Patient Name (Patient Control Number) (ID)": claim.patient.last_first_name + ' (' + claim.patient.identification + ') (' + claim.patient.identification_code + ')',
            "Patient Amt": '',
            "Total Charged Amt": claim.claim.charge_amount,
            "Total Paid Amt": claim.claim.paid_amount,
        }

        return datum

    @classmethod
    def build(cls, file_path: str) -> "TransactionSet":
        """build a TransactionSet from a file path"""
        check_information = None
        interchange = None
        financial_information = None
        claims = []
        organizations = []

        with open(file_path) as f:
            file = f.read()

        segments = file.split("~")
        segments = [segment.strip() for segment in segments]

        segments = iter(segments)
        segment = None

        while True:
            response = cls.build_attribute(segment, segments)
            segment = response.segment
            segments = response.segments

            # no more segments to parse
            if response.segments is None:
                break

            if response.key == "check information":
                check_information = response.value

            if response.key == "interchange":
                interchange = response.value

            if response.key == "financial information":
                financial_information = response.value

            if response.key == "organization":
                organizations.append(response.value)

            if response.key == "claim":
                claims.append(response.value)

        return TransactionSet(
            check_information, interchange, financial_information, claims, organizations, file_path
        )

    @classmethod
    def build_attribute(
        cls, segment: Optional[str], segments: Iterator[str]
    ) -> BuildAttributeResponse:
        """build an attribute from a segment and segments iterator"""
        if segment is None:
            try:
                segment = segments.__next__()
            except StopIteration:
                return BuildAttributeResponse(None, None, None, None)

        identifier = find_identifier(segment)

        if identifier == CheckEFT.identification:
            check_information = CheckEFT(segment)
            return BuildAttributeResponse(
                "check information", check_information, None, segments
            )

        if identifier == InterchangeSegment.identification:
            interchange = InterchangeSegment(segment)
            return BuildAttributeResponse("interchange", interchange, None, segments)

        if identifier == FinancialInformationSegment.identification:
            financial_information = FinancialInformationSegment(segment)
            return BuildAttributeResponse(
                "financial information", financial_information, None, segments
            )

        if identifier == OrganizationLoop.initiating_identifier:
            organization, segments, segment = OrganizationLoop.build(segment, segments)
            return BuildAttributeResponse(
                "organization", organization, segment, segments
            )

        elif identifier == ClaimLoop.initiating_identifier:
            claim, segments, segment = ClaimLoop.build(segment, segments)
            return BuildAttributeResponse("claim", claim, segment, segments)

        else:
            return BuildAttributeResponse(None, None, None, segments)


if __name__ == "__main__":
    pass



""" Location (N4) segment """
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.segments.utilities import split_segment

class Location:
    """Location (N4) segment class"""

    identification = "N4"

    identifier = Identifier()

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.city = segment[1]
        self.state = segment[2]
        self.zip_code = segment[3]

    def __repr__(self):
        return "\n".join(str(item) for item in self.__dict__.items())


if __name__ == "__main__":
    pass



""" Organization Loop """

# Standard library imports
from typing import Iterator, Tuple, Optional

# Local application imports
from rsa_835_parser.segments.organization import Organization as OrganizationSegment
from rsa_835_parser.segments.claim import Claim as ClaimSegment
from rsa_835_parser.segments.address import Address as AddressSegment
from rsa_835_parser.segments.location import Location as LocationSegment
from rsa_835_parser.segments.utilities import find_identifier

class Organization:
    """ Organization Loop class """
    initiating_identifier = OrganizationSegment.identification
    terminating_identifiers = [
        ClaimSegment.identification,
        OrganizationSegment.identification,
        'SE'
    ]

    def __init__(self, organization: OrganizationSegment = None, location: LocationSegment = None,
				 address: AddressSegment = None):
        self.organization = organization
        self.location = location
        self.address = address

    def __repr__(self):
        return '\n'.join(str(item) for item in self.__dict__.items())

    @classmethod
    def build(cls, current_segment: str, segments: Iterator[str]) -> Tuple[
        'OrganizationSegment', Optional[Iterator[str]], Optional[str]
    ]:
        """ Build organization loop """
        organization = Organization()
        organization.organization = OrganizationSegment(current_segment)
        while True:
            try:
                segment = segments.__next__()
                identifier = find_identifier(segment)

                if identifier == AddressSegment.identification:
                    organization.address = AddressSegment(segment)

                elif identifier == LocationSegment.identification:
                    organization.location = LocationSegment(segment)

                elif identifier in cls.terminating_identifiers:
                    return organization, segments, segment

            except StopIteration:
                return organization, None, None


if __name__ == '__main__':
    pass



""" utilities.py - helper functions for parsing 835 segments"""
from typing import List, Optional

def split_segment(segment: str) -> List[str]:
    """different payers use different characters to delineate elements"""
    asterisk = "*"
    pipe = "|"

    asterisk_segment_count = len(segment.split(asterisk))
    pipe_segment_count = len(segment.split(pipe))

    if asterisk_segment_count > pipe_segment_count:
        return segment.split(asterisk)
    else:
        return segment.split(pipe)


def find_identifier(segment) -> str:
    """finds the segment identifier"""
    segment = split_segment(segment)
    return segment[0]


def get_element(segment: List[str], index: int, default=None) -> Optional[str]:
    """gets an element from a segment"""
    element = default
    if index < len(segment):
        element = segment[index]

    return element



""" Entity (NM1) segment """

# Local imports
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.elements.entity_code import EntityCode
from rsa_835_parser.elements.entity_type import EntityType
from rsa_835_parser.elements.identification_code_qualifier import (
    IdentificationCodeQualifier,
)
from rsa_835_parser.segments.utilities import split_segment, get_element

class Entity:
    """ Entity (NM1) segment class """
    identification = 'NM1'

    identifier = Identifier()
    entity = EntityCode()
    type = EntityType()
    identification_code_qualifier = IdentificationCodeQualifier()

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.entity = segment[1]
        self.type = segment[2]
        self.last_name = segment[3]
        self.first_name = get_element(segment, 4)
        self.identification_code_qualifier = get_element(segment, 8)
        self.identification_code = get_element(segment, 9)

    def __repr__(self):
        return '\n'.join(str(item) for item in self.__dict__.items())

    @property
    def name(self) -> str:
        """ Returns the name of the entity """
        return f'{self.first_name} {self.last_name}'.title()
    
    @property
    def last_first_name(self) -> str:
        """ Returns the patient name in format uppercase Last Name, First Name of the entity """
        return f'{self.last_name}, {self.first_name}'.upper()

if __name__ == '__main__':
    pass



""" Amount (AMT) segment """
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.elements.dollars import Dollars
from rsa_835_parser.elements.amount_qualifier import AmountQualifier
from rsa_835_parser.segments.utilities import split_segment

class Amount:
    """ Amount (AMT) segment """
    identification = 'AMT'

    identifier = Identifier()
    qualifier = AmountQualifier()
    amount = Dollars()

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.qualifier = segment[1]
        self.amount = segment[2]

    def __repr__(self):
        return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
    pass



""" Entity Type """
from rsa_835_parser.elements import Element

# https://magnacare.com/wp-content/uploads/pdf/MagnacareCompanionGuide_835_5010A1.pdf
entity_types = {
    "1": "person",
    "2": "entity",
}


class EntityType(Element):
    """Entity Type class"""

    def parser(self, value: str) -> str:
        return entity_types.get(value, value)



""" Address segment """
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.segments.utilities import split_segment


class Address:
    """ Address (N3) segment class """
    identification = 'N3'

    identifier = Identifier()

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.address = segment[1]

    def __repr__(self):
        return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
    pass



""" Adjustment Reason Code """
from rsa_835_parser.elements import Element, Code

# https://x12.org/codes/claim-adjustment-reason-codes
adjustment_reason_codes = {
	'45': 'Charge exceeds fee schedule maximum allowable or contracted/legislated fee arrangement.',
	'243': 'Services not authorized by network/primary care providers.',
	'29': 'The time limit for filing has expired.',
	'251': 'The attachment/other documentation that was received was incomplete or deficient.',
	'2': 'Coinsurnace Amount.',
	'96': 'Non-covered charge(s). See remark code.',
	'3': 'Co-payment Amount.',
	'16': 'Claim/service lacks information or has submission/billing error(s).',
	'B15':'This service/procedure requires that a qualifying service/procedure be received and covered. The qualifying other service/procedure has not been received/adjudicated.',
	'A1': 'Claim/Service denied. See remark code.',
	'1': 'Deductible Amount',
	'4': 'The procedure code is inconsistent with the modifier used. Usage: Refer to the 835 Healthcare Policy Identification Segment (loop 2110 Service Payment Information REF), if present.',
	'18': 'Exact duplicate claim/service (Use only with Group Code OA except where state workers\' compensation regulations requires CO)',
	'23': 'The impact of prior payer(s) adjudication including payments and/or adjustments. (Use only with Group Code OA)',
	'26': 'Expenses incurred prior to coverage.',
	'27': 'Expenses incurred after coverage terminated.',
	'97': 'The benefit for this service is included in the payment/allowance for another service/procedure that has already been adjudicated. Usage: Refer to the 835 Healthcare Policy Identification Segment (loop 2110 Service Payment Information REF), if present.',
	'109': 'Claim/service not covered by this payer/contractor. You must send the claim/service to the correct payer/contractor.',
	'151': 'Payment adjusted because the payer deems the information submitted does not support this many/frequency of services.',
	'234': 'This procedure is not paid separately. At least one Remark Code must be provided (may be comprised of either the NCPDP Reject Reason Code, or Remittance Advice Remark Code that is not an ALERT.)',
	'272': 'Coverage/program guidelines were not met.',
}


class AdjustmentReasonCode(Element):
    """ Adjustment Reason Code class """

    def parser(self, value: str) -> Code:
        description = adjustment_reason_codes.get(value, None)
        return Code(value, description)



""" Claim Segment """

# Local imports
from rsa_835_parser.elements.identifier import Identifier
from rsa_835_parser.elements.claim_status import ClaimStatus
from rsa_835_parser.elements.dollars import Dollars
from rsa_835_parser.segments.utilities import split_segment

class Claim:
    """ Claim (CLP) segment class """
    identification = 'CLP'

    identifier = Identifier()
    status = ClaimStatus()
    charge_amount = Dollars()
    paid_amount = Dollars()

    # B Consulting Group needs to add the following elements to the claim segment: Claim Number
    claim_number: str = None

    def __init__(self, segment: str):
        self.segment = segment
        segment = split_segment(segment)

        self.identifier = segment[0]
        self.marker = segment[1]
        self.status = segment[2]
        self.charge_amount = segment[3]
        self.paid_amount = segment[4]
        self.claim_number = segment[7]

    def __repr__(self):
        return '\n'.join(str(item) for item in self.__dict__.items())


if __name__ == '__main__':
    pass



""" Date Element """
from typing import Union
from datetime import datetime
from warnings import warn

from rsa_835_parser.elements import Element

class Date(Element):
    """ Date class """

    def parser(self, value: str) -> Union[datetime, str]:
        if len(value) == 10:
            year, month, day, minute, second = [int(value[i:i + 2]) for i in range(0, len(value), 2)]
            return datetime(2000 + year, month, day, minute, second)

        elif len(value) == 8:
            year, month, day = int(value[:4]), int(value[4:6]), int(value[6:])
            return datetime(year, month, day)

        else:
            warn(f'Unable to parse {value} into a datetime')
            return value



""" Date Qualifier """
from rsa_835_parser.elements import Element

# https://ediacademy.com/blog/x12-date-time-qualifiers/
date_qualifiers = {
	'050': 'received',
	'150': 'service period start',
	'151': 'service period end',
	'472': 'service',
	'232': 'claim statement period start',
	'233': 'claim statement period end',
}

class DateQualifier(Element):
    """ Date Qualifier class """

    def parser(self, value: str) -> str:
        return date_qualifiers.get(value, value)



""" Reference Qualifier """
from rsa_835_parser.elements import Element, Code

# https://ushik.ahrq.gov/ViewItemDetails?&system=sdo&itemKey=133213000
reference_qualifiers = {
	'6R': 'provider control number',
	'0K': 'policy form identifying number',
	'PQ': 'payee identification',
	'TJ': 'federal taxpayer identification number',
	'LU': 'location number'
}

class ReferenceQualifier(Element):
    """ Reference Qualifier class """

    def parser(self, value: str) -> Code:
        description = reference_qualifiers.get(value, None)
        return Code(value, description)



"""	Identification Code Qualifier """
from rsa_835_parser.elements import Element

# https://ushik.ahrq.gov/dr.ui.drValueDomain_View?system=mdr&ValueDomainID=4933000&CallingRoutine=$CallingRoutine$&OrganizationID=3&RecordOffset=11&Referer=ValueDomain
identification_code_qualifiers = {
	'MI': 'member identification number',
	'C': "insured's changed unique identification number",
	'PC': 'provider commercial number',
	'XX': 'national provider id',

}


class IdentificationCodeQualifier(Element):
    """ Identification Code Qualifier class """

    def parser(self, value: str) -> str:
        return identification_code_qualifiers.get(value, value)



""" Service Qualifier """
from rsa_835_parser.elements import Element
from rsa_835_parser.elements.utilities import split_element

class ServiceQualifier(Element):
    """Service Qualifier (SVC01) element class"""

    def parser(self, value: str) -> str:
        value = split_element(value)
        qualifier, *_ = value
        return qualifier



""" Amount Qualifier """
from rsa_835_parser.elements import Element

# https://ushik.ahrq.gov/ViewItemDetails?system=mdr&itemKey=133081000
amount_qualifiers = {
	'B6': 'allowed - actual',
	'AU': 'coverage amount'
}


class AmountQualifier(Element):
    """ Amount Qualifier class """

    def parser(self, value: str) -> str:
        return amount_qualifiers.get(value, value)
